from .standards.ISO_9613_1_1993 import SOUND_SPEED


def frequency(frequency, velocity_source, velocity_receiver, soundspeed=SOUND_SPEED):
    return (soundspeed + velocity_receiver) / (soundspeed + velocity_source) * frequency

def velocity(f1, f2, c=SOUND_SPEED):
    return c * (f2 - f1) / (f2 + f1)



from abc import ABC, abstractmethod

class Source(ABC):

    @abstractmethod
    def __init__(self):
        super().__init__()

    @abstractmethod
    def get_sample(self, time:float):
        """return the sample value at the specified time instant. Typically an interpolation technique is used here when the source is a sapmled signal.

        :param time: time of the requested signal sample.
        :type time: float
        """





from dynamic_sound.microphones import MicrophoneArray

import numpy as np

import random
from collections import namedtuple
from scipy.spatial.transform import Rotation

class Hedraphone(MicrophoneArray):
    MicPCB = namedtuple('MicPCB', ['mics', 'faces'])
    
    @staticmethod
    def rotate_points(points: np.ndarray, angle_z: float, angle_y: float, angle_x: float) -> np.ndarray:
        R= Rotation.from_euler('zyx', [angle_z, angle_y, angle_x], degrees=True).as_matrix()
        return R @ points

    def __init__(self, file_path, num_external_mics=6, radius_mics=0.012, radius_pcb=0.022, thickness=0.05, sideboard_angle=30, spacing=0.0, rnd_angle=None, rnd_position=None, sample_rate = 48_000, sample_width=4):
        self.pcb = []
        mics = self._generate_mics(num_external_mics, radius_mics)
        faces = self._generate_faces(num_external_mics, radius_pcb, thickness)
        
        # randomize pcb angles
        if rnd_angle is not None:
            angle_x = random.uniform(-rnd_angle, rnd_angle)
            angle_y = random.uniform(-rnd_angle, rnd_angle)
            angle_z = random.uniform(-rnd_angle, rnd_angle)
            mics = self.rotate_points(mics, angle_z, angle_y, angle_x)
            faces = self.rotate_points(faces, angle_z, angle_y, angle_x)

        # randomize pcb positions
        if rnd_position is not None:
            pos_x = random.uniform(-rnd_position, rnd_position)
            pos_y = random.uniform(-rnd_position, rnd_position)
            pos_z = random.uniform(-rnd_position, rnd_position)
            mics = np.array([[pos_x], [pos_y], [pos_z]]) + mics
            faces = np.array([[pos_x], [pos_y], [pos_z]]) + faces
        
        self.pcb.append(self.MicPCB(mics=mics, faces=faces))

        for i in range(num_external_mics):

            mics = self._generate_mics(num_external_mics, radius_mics)
            faces = self._generate_faces(num_external_mics, radius_pcb, thickness)
            
            # randomize pcb angles
            if rnd_angle is not None:
                angle_x = random.uniform(-rnd_angle, rnd_angle)
                angle_y = random.uniform(-rnd_angle, rnd_angle)
                angle_z = random.uniform(-rnd_angle, rnd_angle)
                mics = self.rotate_points(mics, angle_z, angle_y, angle_x)
                faces = self.rotate_points(faces, angle_z, angle_y, angle_x)

            # randomize pcb positions
            if rnd_position is not None:
                pos_x = random.uniform(-rnd_position, rnd_position)
                pos_y = random.uniform(-rnd_position, rnd_position)
                pos_z = random.uniform(-rnd_position, rnd_position)
                mics = np.array([[pos_x], [pos_y], [pos_z]]) + mics
                faces = np.array([[pos_x], [pos_y], [pos_z]]) + faces

            mics = self.rotate_points(mics, angle_z=180.0, angle_y=0.0, angle_x=0.0)
            mics = np.array([[spacing], [0.0], [0.0]]) + mics
            mics = self.rotate_points(mics, angle_z=0.0, angle_y=sideboard_angle, angle_x=0.0)
            mics = np.array([[spacing], [0.0], [0.0]]) + mics
            mics = self.rotate_points(mics, angle_z=i*360.0/num_external_mics, angle_y=0.0, angle_x=0.0)
                
            faces = self.rotate_points(faces, angle_z=180.0, angle_y=0.0, angle_x=0.0)
            faces = np.array([[spacing], [0.0], [0.0]]) + faces
            faces = self.rotate_points(faces, angle_z=0.0, angle_y=sideboard_angle, angle_x=0.0)
            faces = np.array([[spacing], [0.0], [0.0]]) + faces
            faces = self.rotate_points(faces, angle_z=i*360.0/num_external_mics, angle_y=0.0, angle_x=0.0)

            self.pcb.append(self.MicPCB(mics=mics, faces=faces))
        
        positions = np.concatenate([pcb.mics for pcb in self.pcb], axis=1).T
        rotations = np.tile([1, 0, 0, 0], (len(positions), 1))
        super().__init__(file_path, np.hstack((positions, rotations)), sample_rate, sample_width)

    def _generate_mics(self, num_mics, radius):
        mics = [(0.0, 0.0, 0.0)]
        for i in range(num_mics):
            angle = ((2 * np.pi * i) - np.pi) / num_mics
            x = radius * np.cos(angle)
            y = radius * np.sin(angle)
            z = 0.0
            mics.append((x, y, z))
        return np.array(mics).T

    def _generate_faces(self, num_mics, radius, thickness):
        faces = []
        for i in range(num_mics):
            angle = ((2 * np.pi * i) - np.pi) / num_mics
            x = radius * np.cos(angle)
            y = radius * np.sin(angle)
            z = 0.0
            faces.append((x, y, z))
        return np.array(faces).T

    def plot_figure(self, show=True, ax=None):
        import matplotlib.pyplot as plt
        from mpl_toolkits.mplot3d.art3d import Poly3DCollection

        if ax is None:
            fig = plt.figure()
            ax = fig.add_subplot(111, projection='3d')
        
        mics = self.get_microphones().T

        ax.scatter(mics[0], mics[1], mics[2], marker='o')
        for pcb in self.pcb:
            ax.add_collection3d(Poly3DCollection([pcb.faces.T], alpha=0.5, facecolor='cyan', edgecolor='k'))
        ax.set_aspect('equal')

        ax.set_title("HedraPhone")
        ax.set_xlabel("X")
        ax.set_ylabel("Y")
        ax.set_zlabel("Z")
        ax.grid(True)

        if show:
            plt.show()
        return ax


class Hedraphone_v1(Hedraphone):
    def __init__(self, file_path, rnd_angle=None, rnd_position=None, sample_rate=48000, sample_width=4):
        super().__init__(file_path, num_external_mics=5, radius_mics=0.012, radius_pcb=0.022, thickness=0.00155, sideboard_angle=63.43, spacing=0.022, rnd_angle=rnd_angle, rnd_position=rnd_position, sample_rate=sample_rate, sample_width=sample_width)


class Hedraphone_v2(Hedraphone):
    def __init__(self, file_path, rnd_angle=None, rnd_position=None, sample_rate=48000, sample_width=4):
        super().__init__(file_path, num_external_mics=5, radius_mics=0.014, radius_pcb=0.022, thickness=0.00155, sideboard_angle=63.43, spacing=0.022, rnd_angle=rnd_angle, rnd_position=rnd_position, sample_rate=sample_rate, sample_width=sample_width)





from .standards.ISO_9613_1_1993 import SOUND_SPEED
import numpy as np

def speed_of_sound(temperature=20, humidity=50):
    c = 331.3 + 0.606 * temperature + 0.0124 * humidity
    return c


def wavelength(frequency, c=SOUND_SPEED):
    return c / frequency

def angular_frequency(frequency):
    return 2.0 * np.pi * frequency



from ._source import Source

import numpy as np

class WhiteNoise(Source):
    def __init__(self, duration, sample_rate=48_000, amplitude=1.0):
        super().__init__()
        self.sample_rate = sample_rate
        self.signal = amplitude * np.random.uniform(-1.0, 1.0, int(duration * sample_rate))

    def get_sample(self, time:float):
        time_int = int(time * self.sample_rate)
        time_frac = time * self.sample_rate - time_int
        return (1.0 - time_frac) * self.signal[time_int] + time_frac * self.signal[time_int + 1]
    




class Air:
    def __init__(self, temperature, pressure, relative_humidity):
        self.temperature = temperature
        self.pressure = pressure
        self.relative_humidity = relative_humidity
    
    def set(self, temperature=None, pressure=None, relative_humidity=None):
        self.temperature = temperature if temperature is not None else self.temperature
        self.pressure = pressure if pressure is not None else self.pressure
        self.relative_humidity = relative_humidity if relative_humidity is not None else self.relative_humidity



import os
import numpy as np
from scipy.spatial.transform import Rotation, Slerp
from scipy.interpolate import interp1d

class Path:
    def __init__(self, positions=None, *, file=None):
        self.positions = None
        self.duration = 0.0

        if positions is not None:
            self.positions = np.array(positions, dtype=np.float64)
            self.duration = self.positions[-1, 0] - self.positions[0, 0]
        
        if file is not None:
            self.load_path(file)

    def save_path(self, file_path):
        os.makedirs(os.path.dirname(file_path), exist_ok=True)
        np.savetxt(file_path, self.positions, delimiter=',')

    def load_path(self, file_path):
        self.positions = np.genfromtxt(file_path, delimiter=',', dtype=np.float64)
        self.duration = self.positions[-1][0] - self.positions[0][0]
    
    def get_position(self, time):
        # Ensure time is within range
        if self.positions[0, 0] <= time < self.positions[-1, 0]:

            # Find interval containing `time`
            for i in range(len(self.positions) - 1):
                t0 = self.positions[i, 0]
                t1 = self.positions[i + 1, 0]
                if t0 <= time < t1:
                    # Linear interpolation factor
                    alpha = (time - t0) / (t1 - t0)

                    # --- Position interpolation (linear) ---
                    p0 = self.positions[i, 1:4]
                    p1 = self.positions[i + 1, 1:4]
                    position = p0 + alpha * (p1 - p0)

                    # --- Rotation interpolation (slerp) ---
                    q0 = self.positions[i, 4:8]
                    q1 = self.positions[i + 1, 4:8]
                    
                    q0 /= np.linalg.norm(q0)  # Normalize quaternions to avoid numerical issues
                    q1 /= np.linalg.norm(q1)

                    # Create Rotation objects
                    key_rots = Rotation.from_quat([q0, q1], scalar_first=True)
                    key_times = [0, 1]

                    # Create Slerp object
                    slerp = Slerp(key_times, key_rots)

                    # Interpolate
                    interp_rot = slerp([alpha])[0]
                    rotation = interp_rot.as_quat(scalar_first=True)

                    return position, rotation
        
        # If time not in the range
        return None, None

    
    def interpolate_path(self, num_points=50):
        # Split input
        t = self.positions[:, 0]
        pos = self.positions[:, 1:4]
        quat = self.positions[:, 4:8]

        # New uniform time vector
        t_new = np.linspace(t[0], t[-1], num_points)

        # --- Position interpolation ---
        interp_pos = np.zeros((num_points, 3))
        for i in range(3):
            f = interp1d(t, pos[:, i], kind='cubic')
            interp_pos[:, i] = f(t_new)

        # --- Quaternion interpolation ---
        rotations = Rotation.from_quat(quat, scalar_first=True)
        slerp = Slerp(t, rotations)
        interp_rot = slerp(t_new)
        interp_quat = interp_rot.as_quat(scalar_first=True)

        # Combine interpolated data
        self.positions = np.column_stack([t_new, interp_pos, interp_quat])

    def plot_path_3d(self, show=True, ax=None, dot_every=1, legend=True):
        import matplotlib.pyplot as plt

        pos = self.positions[:, 1:4]

        if ax is None:
            fig = plt.figure()
            ax = fig.add_subplot(111, projection='3d')

        # Path line
        ax.plot(pos[:, 0], pos[:, 1], pos[:, 2], 'b-', linewidth=1.5)

        # dots
        ax.scatter(
            pos[::dot_every, 0],
            pos[::dot_every, 1],
            pos[::dot_every, 2],
            color='blue',
            s=10,
            alpha=0.6
        )

        # Start and end points
        ax.scatter(pos[0, 0], pos[0, 1], pos[0, 2], color='green', s=60, label='Start', zorder=3)
        ax.scatter(pos[-1, 0], pos[-1, 1], pos[-1, 2], color='red', s=60, label='End', zorder=3)
        
        # Labels and aesthetics
        ax.set_xlabel("x")
        ax.set_ylabel("y")
        ax.set_zlabel("z")
        ax.grid(True)

        if legend:
            # Shrink current axis's height by 10% on the bottom
            box = ax.get_position()
            ax.set_position([box.x0, box.y0 + box.height * 0.1,
                             box.width, box.height * 0.9])
            
            # Put a legend below current axis
            ax.legend(loc='upper center', bbox_to_anchor=(0.5, -0.05),
                      fancybox=True, shadow=True, ncol=5)

        if show:
            plt.show()
        return ax


    def plot_quaternion_directions(self, show=True, ax=None, step=1, scale=0.1, legend=True):
        import matplotlib.pyplot as plt

        pos = self.positions[:, 1:4]
        quat = self.positions[:, 4:8]
        rot = Rotation.from_quat(quat, scalar_first=True)

        # Subsample
        idx = np.arange(0, len(pos), step)
        p = pos[idx]
        Rm = rot[idx].as_matrix()  # shape (m, 3, 3)

        if ax is None:
            fig = plt.figure()
            ax = fig.add_subplot(111, projection='3d')

        # Plot axes for each sampled frame
        for i in range(len(p)):
            origin = p[i]
            R = Rm[i]
            # Each column of R is a local axis in world coordinates
            x_dir, y_dir, z_dir = R[:, 0], R[:, 1], R[:, 2]

            ax.quiver(*origin, *x_dir, length=scale, normalize=True, color='r', linewidth=1)
            ax.quiver(*origin, *y_dir, length=scale, normalize=True, color='g', linewidth=1)
            ax.quiver(*origin, *z_dir, length=scale, normalize=True, color='b', linewidth=1)

        ax.set_xlabel("x")
        ax.set_ylabel("y")
        ax.set_zlabel("z")
        ax.grid(True)
        ax.set_box_aspect([1, 1, 1])

        if legend:
            ax.plot([], [], [], color='r', label='x')
            ax.plot([], [], [], color='g', label='y')
            ax.plot([], [], [], color='b', label='z')
        
            # Shrink current axis's height by 10% on the bottom
            box = ax.get_position()
            ax.set_position([box.x0, box.y0 + box.height * 0.1,
                             box.width, box.height * 0.9])
            
            # Put a legend below current axis
            ax.legend(loc='upper center', bbox_to_anchor=(0.5, -0.05),
                      fancybox=True, shadow=True, ncol=5)

        if show:
            plt.show()
        return ax



import numpy as np

class MicrophoneArray:
    def __init__(self, file_path:str, positions:list, sample_rate:int=48_000, sample_width=4):
        self.positions = np.array(positions, dtype=np.float64)
        self.num_channels = len(self.positions)
        self.sample_width = sample_width
        self.file_path = file_path
        self.sample_rate = sample_rate

    def get_microphones(self) -> np.ndarray:
        return self.positions

class Microphone(MicrophoneArray):
    def __init__(self, file_path:str, sample_rate:int=48_000, sample_width=4):
        position = [[0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0], ]
        super().__init__(file_path, position, sample_rate, sample_width)



import os
from tqdm import tqdm
import numpy as np
import wave
from collections import deque
from scipy.signal import firwin2, lfilter
import struct

from .environment import Air
from .acoustics import attenuations
from .acoustics.standards.ISO_9613_1_1993 import sound_speed, attenuation_coefficients, REFERENCE_TEMPERATURE, SOUND_SPEED
from .environment import Path
from .sources import Source
from .microphones import MicrophoneArray
from scipy.spatial.transform import Rotation as R

class Simulation:
    def __init__(self, temperature=20, pressure=1, relative_humidity=50):
        self.air = Air(temperature=temperature, pressure=pressure, relative_humidity=relative_humidity)
        self._microphones = []
        self._sources = []

    def add_microphone(self, path:Path, microphone: MicrophoneArray):
        self._microphones.append((path, microphone))

    def add_source(self, path:Path, source:Source):
        self._sources.append((path, source))

    @staticmethod
    def _compute_emission(position_receiver, time_receiver, source_path, c=SOUND_SPEED):
        for path_index in range(len(source_path.positions) - 1):
            t0 = source_path.positions[path_index, 0]
            t1 = source_path.positions[path_index+1, 0]
            p0 = source_path.positions[path_index, 1:4]
            p1 = source_path.positions[path_index+1, 1:4]
            v = (p1 - p0) / (t1 - t0)

            if t0 <= time_receiver:
                d0 = position_receiver - p0

                A = np.dot(v, v) - c**2
                B = 2 * (c**2 * (time_receiver - t0) - np.dot(d0, v))
                C = np.dot(d0, d0) - (c * (time_receiver - t0))**2

                if A == 0:
                    if B == 0:
                        if C == 0:
                            print("ERROR: infinite values")
                            return None, None
                    else:
                        time_emission = (-C / B) + t0
                        if t0 <= time_emission < t1 and time_emission <= time_receiver:
                            position_emission = p0 + (v * (time_emission - t0))
                            return time_emission, position_emission
                else:
                    delta = B**2 - 4*A*C
                    if delta > 0:
                        sqrt_delta = np.sqrt(delta)
                        time_emission = min((-B - sqrt_delta) / (2*A) + t0, (-B + sqrt_delta) / (2*A) + t0)
                        if t0 <= time_emission < t1 and time_emission <= time_receiver:
                            position_emission = p0 + (v * (time_emission - t0))
                            return time_emission, position_emission
        return None, None

    def run(self):
        c = sound_speed(temperature=self.air.temperature+273.15)

        for microphone_path, microphone in self._microphones:
            dst_path = os.path.dirname(microphone.file_path)
            if dst_path:
                os.makedirs(dst_path, exist_ok=True)

            with wave.open(microphone.file_path, mode="wb") as wave_file:
                wave_file.setnchannels(microphone.num_channels)
                wave_file.setsampwidth(microphone.sample_width)
                wave_file.setframerate(microphone.sample_rate)

                # air absorption filter
                filter_len = 513
                frequencies = np.linspace(0, microphone.sample_rate/2, num=512)
                air_absorption_coefficients = attenuation_coefficients(
                    frequency=frequencies,
                    temperature=self.air.temperature + 273.15,
                    relative_humidity=self.air.relative_humidity,
                    pressure=self.air.pressure * 101.325
                )

                out_buffer = [deque(np.zeros(filter_len), maxlen=filter_len) for _ in range(microphone.num_channels)]
                out_samples = np.zeros((int(microphone.sample_rate * microphone_path.duration), microphone.num_channels))

                for sample_index, time_receiver in tqdm([(index, index/microphone.sample_rate) for index in range(int(microphone.sample_rate * microphone_path.duration))]):
                    position_array, rotation_array = microphone_path.get_position(time_receiver)

                    for channel_index, microphone_position in enumerate(microphone.get_microphones()):
                        for source_path, source in self._sources:

                            rotation_receiver = R.from_quat(rotation_array, scalar_first=True)
                            position_receiver = position_array + rotation_receiver.apply(microphone_position[0:3])

                            time_emission, position_emission = self._compute_emission(position_receiver=position_receiver, time_receiver=time_receiver, source_path=source_path, c=c)

                            if time_emission is not None:
                                distance = np.linalg.norm(position_receiver - position_emission)

                                attenuation_geom = attenuations.geometric(distance)
                                air_coeff = 10 ** (-air_absorption_coefficients * distance / 20.0)  # Convert coeffs in dB to linear scale
                                air_fir_coefficients = firwin2(filter_len, frequencies, air_coeff, fs=microphone.sample_rate)

                                out_buffer[channel_index].appendleft(source.get_sample(time_emission) * attenuation_geom)
                                out_samples[sample_index, channel_index] += air_fir_coefficients.dot(out_buffer[channel_index])
                
                interleaved = (np.clip(out_samples, -1.0, 1.0) * np.iinfo(np.int32).max).astype(np.int32).reshape(-1)
                wave_file.writeframes(struct.pack("<" + "i" * len(interleaved), *interleaved)) # int32 (little-endian)



import numpy as np

SOUND_SPEED = 343.2  # [m/s] = 1235.52 km/h
REFERENCE_TEMPERATURE = 293.15  # [K] = 20.0 °C
REFERENCE_PRESSURE = 101.325  # [kPa] = 1 atm
TRIPLE_TEMPERATURE = 273.16  # [K] = 0.01 °C


def sound_speed(temperature):
    return SOUND_SPEED * np.sqrt(temperature / REFERENCE_TEMPERATURE)

def saturation_pressure(temperature, reference_pressure=REFERENCE_PRESSURE, triple_temperature=TRIPLE_TEMPERATURE):  # (ISO 9613-1: B.2, B.3)
    return reference_pressure * 10.0**(-6.8346 * (triple_temperature / temperature)**(1.261) + 4.6151)

def molar_concentration_water_vapour(relative_humidity, saturation_pressure, pressure): # # (ISO 9613-1: B.1)
    return relative_humidity * saturation_pressure / pressure

def relaxation_frequency_oxygen(pressure, h, reference_pressure=REFERENCE_PRESSURE):  # (ISO 9613-1: 3)
    return pressure / reference_pressure * (24.0 + 4.04 * 10.0**4.0 * h * (0.02 + h) / (0.391 + h))

def relaxation_frequency_nitrogen(pressure, temperature, h, reference_temperature=REFERENCE_TEMPERATURE, reference_pressure=REFERENCE_PRESSURE):  # (ISO 9613-1: 4)
    return pressure / reference_pressure * (temperature / reference_temperature)**(-0.5) * (9.0 + 280.0 * h * np.exp(-4.170 * ((temperature / reference_temperature)**(-1.0 / 3.0) - 1.0)))

def attenuation_coefficients(frequency, temperature, relative_humidity, pressure, reference_temperature=REFERENCE_TEMPERATURE, reference_pressure=REFERENCE_PRESSURE, triple_temperature=TRIPLE_TEMPERATURE):  # (ISO 9613-1: 5)   
    Psat = saturation_pressure(temperature=temperature, reference_pressure=reference_pressure, triple_temperature=triple_temperature)
    h = molar_concentration_water_vapour(relative_humidity=relative_humidity, saturation_pressure=Psat, pressure=pressure)
    frO = relaxation_frequency_oxygen(pressure=pressure, h=h, reference_pressure=reference_pressure)
    frN = relaxation_frequency_nitrogen(pressure, temperature, h, reference_temperature=reference_temperature, reference_pressure=reference_pressure)

    return 8.686 * frequency**2 * (
        (1.84e-11 * (pressure / reference_pressure)**(-1.0) * (temperature / reference_temperature)**0.5) +
        (temperature / reference_temperature)**(-5.0/2.0) * (
            0.01275 * np.exp(-2239.1 / temperature) / (frO + (frequency**2 / frO)) +
            0.1068 * np.exp(-3352.0 / temperature) / (frN + (frequency**2 / frN)) 
        )
    )

if __name__ == "__main__":
    print(sound_speed(273.15+20))





from ._source import Source

import librosa
import numpy as np

class AudioFile(Source):
    def __init__(self, filename, sample_rate=None, gain_db=0.0, loop=True):
        super().__init__()
        self.signal, self.sample_rate = librosa.load(filename, sr=sample_rate, mono=True)
        self.loop = loop
        self.length = len(self.signal)
        if gain_db != 0.0:
            self.signal = self.signal * 10**(gain_db / 20.0)

    def get_sample(self, time:float):
        time_int = int(time * self.sample_rate)
        if self.loop == False and time_int >= self.length-1:
            return 0.0
        
        time_frac = time * self.sample_rate - time_int
        return (1.0 - time_frac) * self.signal[time_int % self.length] + time_frac * self.signal[(time_int + 1) % self.length]




from ._source import Source

import numpy as np

class SineWave(Source):
    def __init__(self, frequency, amplitude=1.0):
        super().__init__()
        self.amplitude = amplitude
        self.frequency = frequency

    def get_sample(self, time:float):
        return self.amplitude * np.sin(2 * np.pi * self.frequency * time)

